#include "gl_viewer.h"

#include <cassert>
#include <iostream>

using namespace std;

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>

#include "gl_camera.h"

gl_viewer* gl_viewer::singleton = NULL; 

gl_viewer::gl_viewer()
{
    assert(!singleton); // ensure only one instance is created
    singleton = this;
}

gl_viewer::~gl_viewer()
{
    singleton = NULL;
}

void gl_viewer::init(int argc, char *argv[], int width, int height)
{
    // set up glut
    glutInit(&argc, argv);

    // create a 24-bit double-buffered window with a z-buffer
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH); 
    glutInitWindowSize(width, height);
    
    glutCreateWindow("CS130 3D Viewer");

    // called when Glut needs to display
    glutDisplayFunc(&gl_viewer::glut_display_event_wrapper);

    // called when Glut has mouse input
    glutMouseFunc(&gl_viewer::glut_mouse_event_wrapper);

    // called when Glut has key input
    glutKeyboardFunc(&gl_viewer::glut_keyboard_event_wrapper);

    // called when the window is resized
    glutReshapeFunc(&gl_viewer::glut_reshape_event_wrapper);
    
    // called when GLUT has nothing to do
    glutIdleFunc(&gl_viewer::glut_display_event_wrapper);
    
    // called when a mouse button is pressed and the mouse is moving
    glutMotionFunc(&gl_viewer::glut_mouse_move_event_wrapper);

    // clear our background to black when glClear is called
    glClearColor(0.0, 0.0, 0.0, 0);
    
    glClearDepth(1.0);

    // enable depth buffer
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);

    const float ambient[] = { 0.5, 0.5, 0.5, 1.0 };
    const float diffuse[] = { 0.4, 0.4, 0.4, 1.0 };
    const float specular[] = { 0.2, 0.2, 0.2, 1.0 };
    const float light_pos[] = { 20.0, 20.0, 20.0, 1,0 };   

    // enable material properties
    glEnable( GL_COLOR_MATERIAL );
    
    // set up 1 light
    glLightfv( GL_LIGHT1, GL_AMBIENT, ambient );
    glLightfv( GL_LIGHT1, GL_DIFFUSE, diffuse );
    glLightfv( GL_LIGHT1, GL_SPECULAR, specular );
    glLightfv( GL_LIGHT1, GL_POSITION, light_pos );
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT1);
    
    // enable smooth shading
    glShadeModel(GL_SMOOTH);

    this->width = width;
    this->height = height;
    this->mouse_last_x = 0;
    this->mouse_last_y = 0;
    this->first_click = true;

}

void gl_viewer::run()
{
    // pass execution to GLUT. Now GLUT is running the main loop.
    glutMainLoop();
}


void gl_viewer::glut_display_event_wrapper()
{
    if (!singleton)
        return;

    // clear our back color buffer and depth buffer
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glLoadIdentity();
    
    // apply our camera transformation
    singleton->camera.apply_gl_transform();
    
    glDisable(GL_LIGHTING);
   
    /*
     * render axes
     */
    glLineWidth(3.0);
    glPushMatrix();

    glScaled(0.5, 0.5, 0.5); 
    glBegin(GL_LINES);
    
    // draw the x-axis red
    glColor3f(1.0, 0.0, 0.0);
    glVertex3f(0.0, 0.0, 0.0);
    glVertex3f(1.0, 0.0, 0.0);
    
    // draw the y-axis green
    glColor3f(0.0, 1.0, 0.0);
    glVertex3f(0.0, 0.0, 0.0);
    glVertex3f(0.0, 1.0, 0.0);
    
    // draw the z-axis blue
    glColor3f(0.0, 0.0, 1.0);
    glVertex3f(0.0, 0.0, 0.0);
    glVertex3f(0.0, 0.0, 1.0);
    
    glEnd();
    glPopMatrix();
    
    /*
     * render grid
     */
    glLineWidth(1.0);
    glColor3f(.20,.20,.20);
    
    glBegin(GL_LINES);

    for (int i= 0; i <=200; i++)
    {
        int k = -100;
        k +=i;
        glVertex3f(100,0,k);
        glVertex3f(-100,0,k);
    }
    for (int i = 0; i <= 200; i++)
    {
        int k = -100;

        k += i;

        glVertex3f(k,0,100);
        glVertex3f(k,0,-100);
    }
    glEnd();
    
    glEnable(GL_LIGHTING);
    
    // call our gl_viewers draw event and render to back buffer
    singleton->draw_event();

    // swap the back buffer with the front (user always sees front
    // buffer on display)
    glutSwapBuffers();
}



void gl_viewer::glut_mouse_event_wrapper(
    int button, int state, int x, int y
    )
{
    if (!singleton)
        return;
   
    double delta_x, delta_y;

    if (singleton->first_click)
    {
        delta_x = 0.0;
        delta_y = 0.0;

        singleton->first_click = false;
    }
    else
    {
        delta_x = x - singleton->mouse_last_x;
        delta_y = y - singleton->mouse_last_y;
    }

    if (state == gl_viewer::BUTTON_DOWN)
    {
        if (button == gl_viewer::LEFT_BUTTON)
            singleton->mode = 4;
        else if (button == gl_viewer::RIGHT_BUTTON)
            singleton->mode = 2;
        else if (button == gl_viewer::MIDDLE_BUTTON)
            singleton->mode = 6;
    }

    // normally (0,0) is at the top left of screen. Since this is
    // somewhat unintuitive, it has been changed to be the bottom
    // left
    singleton->mouse_event(
        (mouse_button)button, (mouse_button_state)state, 
        x, singleton->height - y
    );
    
    singleton->mouse_last_x = x;
    singleton->mouse_last_y = y;
    
}

void gl_viewer::glut_keyboard_event_wrapper(unsigned char key, int x, int y)
{
    if (!singleton)
        return;

    singleton->keyboard_event(key, x, y);
}

void gl_viewer::glut_reshape_event_wrapper(int width, int height)
{
    if (!singleton)
        return;

    singleton->width = width;
    singleton->height = height;

    // adjust the view frustrum and viewport 
    // to reflect the new window dimensions
    glViewport(0, 0, width, height);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    if (height == 0)
        height = 1;

    gluPerspective(45.0f, width / (double)height, 0.1f, 100.0f);

    glMatrixMode(GL_MODELVIEW);
}

void gl_viewer::glut_mouse_move_event_wrapper(int x, int y)
{
    if (!singleton)
        return;

    int delta_x = x - singleton->mouse_last_x;
    int delta_y = y - singleton->mouse_last_y;

    if (singleton->mode == 0)
    {
        singleton->camera.set_delta_elevation(delta_y / 4.0);
    }
    else if (singleton->mode == 1)
    {
        singleton->camera.set_delta_twist(delta_x / 4.0);
    }
    else if (singleton->mode == 2)
    {
        singleton->camera.set_delta_distance(delta_y / 100.0); 
    }
    else if (singleton->mode == 4)
    {
       singleton->delta_x_total = abs(delta_x);
       singleton->delta_y_total = abs(delta_y); 
       singleton->num_motion_calls_thus_far = 0;

       singleton->mode = 5;
    }
    else if (singleton->mode == 5)
    {
       singleton->delta_x_total += abs(delta_x);
       singleton->delta_y_total += abs(delta_y); 
       
       singleton->num_motion_calls_thus_far += 1;
       
       if (singleton->num_motion_calls_thus_far > 3)
        {
            if (singleton->delta_x_total > singleton->delta_y_total)
                singleton->mode = 1;
            else
                singleton->mode = 0;
        }
    }

    singleton->mouse_move_event(x, y);

    singleton->mouse_last_x = x;
    singleton->mouse_last_y = y;
}
